"""parsers [`Module`].

This module contains lawyer's argument parser classes for creating
individual and grouped arguments (`singlet` and `singlets` repectively).

The `singlet` object represents a single argument object whereas the `singlets` object
as the name suggests, is a collection of `singlet` objects or other nested `singlets` objects.
The `singlet` object supports an argument pool called `reference` parameter which is by default
set to `sys.argv` but can be changed to any iterable of strings making it versatile and extensible.
The `singlet` object implements partial parsing during object creation and information such as
argument's presence in the reference, index, notation and as such.

The `singlets` object acts as a highly customized linked list data structure that efficiently
manages nested argument group structures and their help text generation automatically.

Both `singlet` and `singlets` objects support accessible attributes using square brackets.

----------------------------------------
This module is under license as follows:
----------------------------------------

MIT License
-----------

Copyright (c) 2025 Soumyo Deep Gupta

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
---------------------------- END of LICENSE ----------------------------------
"""

from typing import Literal, Generic, TypeVar, Tuple, Iterable
from typing import Any, NoReturn, TypeGuard, Union, Dict, Self
from typing import overload
import sys


default = TypeVar('default')


class singlet(Generic[default]):

    """A class to represent a single argument.
    
    This class contains parameters for a single argument to make it self sufficient.
    Creating the object itself partially parses it and performs conditional checks
    placed to determine validity. It also contains internal mechanisms to fully
    parse this particular argument and return the value(s) as intended.

    Using `str` function on this class will return -> `singlet(*notations)`.

    Few of the attributes of this class are accessible using square brackets and are
    listed in `singlet.__attrs__`. Changing any attribute after creating an instance
    will automatically call the partial parsing method again.
    """


    __attrs__: Tuple[Literal['name'], Literal['notations'], Literal['capture-type'], Literal['helptext'], Literal['default'], Literal['required'], Literal['reference'],
                     Literal['auto-help'], Literal['help-notations'], Literal['application-name'], Literal['application-version'], Literal['application-description'],
                     Literal['present'], Literal['notation'], Literal['notpresent'], Literal['index']]


    __allowed_capture_types__: Tuple[Literal['auto'], Literal['single'], Literal['boolean']]



    @overload
    def __getitem__(self, name: Literal['name'], /) -> str: ...
    @overload
    def __getitem__(self, name: Literal['notations'], /) -> Iterable[str]: ...
    @overload
    def __getitem__(self, name: Literal['capture-type'], /) -> Literal['auto', 'single', 'boolean']: ...
    @overload
    def __getitem__(self, name: Literal['helptext'], /) -> str: ...
    @overload
    def __getitem__(self, name: Literal['default'], /) -> default: ...
    @overload
    def __getitem__(self, name: Literal['required'], /) -> bool: ...
    @overload
    def __getitem__(self, name: Literal['reference'], /) -> Iterable[str]: ...
    @overload
    def __getitem__(self, name: Literal['auto-help'], /) -> bool: ...
    @overload
    def __getitem__(self, name: Literal['help-notations'], /) -> Iterable[str]: ...
    @overload
    def __getitem__(self, name: Literal['application-name'], /) -> str: ...
    @overload
    def __getitem__(self, name: Literal['application-version'], /) -> str: ...
    @overload
    def __getitem__(self, name: Literal['application-description'], /) -> str: ...
    @overload
    def __getitem__(self, name: Literal['present'], /) -> bool: ...
    @overload
    def __getitem__(self, name: Literal['notpresent'], /) -> bool: ...
    @overload
    def __getitem__(self, name: Literal['notation'], /) -> str: ...
    @overload
    def __getitem__(self, name: Literal['index'], /) -> int: ...
    @overload
    def __getitem__(self, name: Any, /) -> NoReturn: ...


    @overload
    def __setitem__(self, name: Literal['name'], value: str, /) -> None: ...
    @overload
    def __setitem__(self, name: Literal['notations'], value: Iterable[str], /) -> None: ...
    @overload
    def __setitem__(self, name: Literal['capture-type'], value: Literal['auto', 'single', 'boolean'], /) -> None: ...
    @overload
    def __setitem__(self, name: Literal['helptext'], value: str, /) -> None: ...
    @overload
    def __setitem__(self, name: Literal['default'], value: default, /) -> None: ...
    @overload
    def __setitem__(self, name: Literal['required'], value: bool, /) -> None: ...
    @overload
    def __setitem__(self, name: Literal['reference'], value: Iterable[str], /) -> None: ...
    @overload
    def __setitem__(self, name: Literal['auto-help'], value: bool, /) -> None: ...
    @overload
    def __setitem__(self, name: Literal['help-notations'], value: Iterable[str], /) -> None: ...
    @overload
    def __setitem__(self, name: Literal['application-name'], value: str, /) -> None: ...
    @overload
    def __setitem__(self, name: Literal['application-version'], value: str, /) -> None: ...
    @overload
    def __setitem__(self, name: Literal['application-description'], value: str, /) -> None: ...
    @overload
    def __setitem__(self, name: str, value: Any, /) -> NoReturn: ...


    def __delitem__(self, name: str, /) -> NoReturn: ...


    def __str__(self) -> str: ...
    def __repr__(self) -> str: ...


    @staticmethod
    def __version_format_checker__(candidate) -> TypeGuard[str]:
        """Checks a candidate for valid version format such as 1.0, 1.2.0, etc."""

    @staticmethod
    def __value_in_reference__(*reference: Union[singlet, singlets], value: str) -> bool:
        """Checks if a value is defined as a notation for any of the references."""

    @staticmethod
    def __indices_of_reference__(*reference: Union[singlet, singlets], index_to_check: int) -> Tuple[int, ...]:
        """Returns a tuple of indices for reference that is present and index of that reference is greater than the index
        to check (index in context of self.reference)."""

    def __init__(self, name: str, *notations: str, capture: Literal['auto', 'single', 'boolean'] = 'auto', helptext: str = ..., required: bool = False, default: default = ...,
                 reference: Iterable[str] = sys.argv, skip: int = 1, auto_help: bool = False, help_notations: Iterable[str] = ..., application_name: str = sys.argv[0], application_version: str = '1.0', application_description: str = ...) -> None:
        """Create a singlet object with given parameters.
        
        This singlet object created will solely base itself on the given parameters and will work by strictly
        abiding to the parameter specifications.

        Once created the attributes that can be modified using square brackets, will have type hints available.

        This class somewhat partially parses the argument upon object creation using the `__partial__` method
        and calls it again when any attribute is changed using square brackets.

        The `reference` parameter specifies the argument pool which will processed to extract parsing information.
        By default its the argument passed to the script (`sys.argv`) and `skip` parameter specifies the number of
        elements of `reference` to skip. In default case, `sys.argv[0]` is the script name and is not required,
        therefore, `skip` is `1`.

        The `default` and `required` parameters are optional and if both are present, it will override the
        `required` functionality because if there is a default value, then why does it need to be marked as
        required?

        ## Singlet Object Types:

        ### 1. A singlet argument that takes a single value.

        Example:

        ```zsh
        $ your-script.py --argument value
        ```

        Here the `--argument` takes a value `value` and to create this configuration,
        use the following code structure:

        ```python
        single_capture_argument = singlet('argument-1', '--argument', '-a', capture='single', ...)
        ```
    
        The `capture` parameter ensures only one value is captured when using `single` as it's value.
        If you are expecting indefinite values for an argument use `capture='auto'`.

        NOTE: This option requires for extra references while parsing, which is all other arguments
        or argument groups that are defined in the current context.

        For example,

        ```python
        argument1 = singlet(..., capture='single', ...)
        argument2 = singlet(...)
        argument3 = singlet(...)

        # to parse capture='single' and detect potential errors, a
        # reference is needed.
        value = argument1.__parse__(argument1, argument2, argument3) # and so on.
        ```

        All arguments (`singlet`) and argument groups (`singlets`) must be passed to the `__parse__`
        method as reference for it to detect potential issues and parse the arguments correctly.
        
        ### 2. A singlet argument that takes indefinite number of values.

        Example:

        ```zsh
        $ your-script.py --argument value1, value2, v3, ... # indefinite number of arguments
        ```

        Here, the `--argument` takes indefinite number of values and to create this configuration,
        use `capture='auto'`. Similar to `capture='single'` this will require a context reference
        (all created `singlet` and `singlets` object) during parsing. (see 1.)

        ### 3. A singlet argument that takes no values but its presence is stored as a boolean parameter.

        Example:
        
        ```zsh
        $ your-script.py --argument # signifies True if present else False
        ```

        To create this configuration, use `capture='boolean'` and parsing it does not require any context
        reference. The `default` value for this type of argument is automatically set to `False` unless
        specified otherwise.
        """

    @overload
    def __helptext__(self, exit: Literal[True], exit_code: int = ...) -> NoReturn:
        """print help text and exit with `exit_code`."""

    @overload
    def __helptext__(self, exit: Literal[False]) -> None:
        """print help text."""

    def __partial__(self, raise_error: bool = True) -> None:
        """Partially parse the argument to retreive information about it from the set reference such as its
        presence and location.

        If `raise_error` is `False`, it will avoid raising any errors (not recommended).
        """

    @overload
    def __parse__(self, *reference: Union[singlet, singlets], raise_error: bool = True) -> Union[Tuple[str, ...], str, bool, default, None]:
        """Fully parse the argument and return the result if any.
        
        Return Type is based on `capture-type` attribute:
        - `capture-type='boolean'` returns bool.
        - `capture-type='single'` returns either str or default or None.
        - `capture-type='auto'` returns str (if only one value), Tuple[str, ..] (if multiple value), default or None.
        """

    @overload
    def __parse__(self, *reference: Union[singlet, singlets], capture_type: Literal['single'], raise_error: bool = True) -> Union[str, default, None]:
        """Fully parse the argument with given `capture_type='single'`"""
    @overload
    def __parse__(self, *reference: Union[singlet, singlets], capture_type: Literal['boolean'], raise_error: bool = True) -> bool:
        """Fully parse the argument with given `capture_type='boolean'`"""
    @overload
    def __parse__(self, *reference: Union[singlet, singlets], capture_type: Literal['auto'], raise_error: bool = True) -> Union[str, default, Tuple[str, ...], None]:
        """Fully parse the argument with given `capture_type='auto'`.
        
        - Returns str if one value is present, else a tuple.
        - Returns default value if the argument is not present.
        - Else None.
        """

    @property
    def __present__(self) -> bool:
        """`True` if singlet argument is present in the argument pool, else `False`."""

    @property
    def __notpresent__(self) -> bool:
        """`True` if singlet argument is not present in the argument pool, else `False`."""

    @property
    def __index__(self) -> int:
        """Index of the singlet argument in the argument pool. If not present, `-1`."""

    @property
    def __notation__(self) -> str:
        """Notation variant of the singlet argument that is found in the argument pool. If not present, `''`."""

    @property
    def __usage__(self) -> str:
        """formatted usage string for this argument."""



class singlets:
    """A class to represent a collection of `singlet` arguments.
    
    The structure for creating `singlets`:

    - A base argument (that identifies this collection)
        - Any number of `singlet` arguments.
        - Any number of `singlets` collections.

    This class contains parameters for a collection of arguments such that a base argument that uniquely
    identifies this collection must be present.
    
    If that base argument is a required argument, then it will act as a simple `singlet` argument being required
    and will not affect any other arguments or colelctions withing this collection. If any `singlet` argument is
    set as required, it will behave different than the base argument: the required `singlet` argument will be
    required only if the base argument (this collection) is being used. Unless the base argument is present, any
    of the other arguments present in this collection will not raise errors for requirement.

    Similarly, unles the base argument is being used, using any other arguments under this collection will raise
    an exception.

    These same set of rules will also be followed by any nested collections under this collection. If a child
    collection's base argument is used but it's parent's base argument is not used, that will also raise an error.
    """


    __attrs__: Tuple[Literal['name'], Literal['parent'], Literal['helptext'], Literal['reference'], Literal['auto-help'], Literal['help-notations'], Literal['application-name'],
                     Literal['application-version'], Literal['application-description'], Literal['collection'], Literal['present'], Literal['notpresent'], Literal['base']]


    @overload
    def __getitem__(self, name: Literal['name'], /) -> str: ...
    @overload
    def __getitem__(self, name: Literal['parent'], /) -> Union['singlets', None]: ...
    @overload
    def __getitem__(self, name: Literal['helptext'], /) -> str: ...
    @overload
    def __getitem__(self, name: Literal['reference'], /) -> Iterable[str]: ...
    @overload
    def __getitem__(self, name: Literal['auto-help'], /) -> bool: ...
    @overload
    def __getitem__(self, name: Literal['help-notations'], /) -> Iterable[str]: ...
    @overload
    def __getitem__(self, name: Literal['application-name'], /) -> str: ...
    @overload
    def __getitem__(self, name: Literal['application-version'], /) -> str: ...
    @overload
    def __getitem__(self, name: Literal['application-description'], /) -> str: ...
    @overload
    def __getitem__(self, name: Literal['collection'], /) -> Dict[str, Union[singlet, 'singlets']]: ...
    @overload
    def __getitem__(self, name: Literal['present'], /) -> bool: ...
    @overload
    def __getitem__(self, name: Literal['notpresent'], /) -> bool: ...
    @overload
    def __getitem__(self, name: Literal['base'], /) -> Union[singlet, None]: ...


    @overload
    def __setitem__(self, name: Literal['name'], value: str, /) -> None: ...
    @overload
    def __setitem__(self, name: Literal['parent'], value: singlets, /) -> None: ...
    @overload
    def __setitem__(self, name: Literal['helptext'], value: str, /) -> None: ...
    @overload
    def __setitem__(self, name: Literal['reference'], value: Iterable[str], /) -> None: ...
    @overload
    def __setitem__(self, name: Literal['auto-help'], value: bool, /) -> None: ...
    @overload
    def __setitem__(self, name: Literal['help-notations'], value: Iterable[str], /) -> None: ...
    @overload
    def __setitem__(self, name: Literal['application-name'], value: str, /) -> None: ...
    @overload
    def __setitem__(self, name: Literal['application-version'], value: str, /) -> None: ...
    @overload
    def __setitem__(self, name: Literal['application-description'], value: str, /) -> None: ...
    @overload
    def __setitem__(self, name: Literal['base'], value: Union[singlet, None], /) -> None: ...
    @overload
    def __setitem__(self, name: Any, value: Any, /) -> NoReturn: ...


    def __delitem__(self, name: Any, /) -> NoReturn: ...


    @property
    def __base__(self) -> Union[singlet, None]:
        """The base argument object for this singlets object if present, else None."""


    def __init__(self, name: str, parent: Union[singlets, None] = ..., helptext: str = ..., reference: Iterable[str] = sys.argv, skip: int = 1, auto_help: bool = False,
                 help_notations: Iterable[str] = ..., application_name: str = sys.argv[0], application_version: str = '1.0', application_description: str = ...) -> None:
        """Create a singlets object with given parameters.
        
        The singlets object is collection of objects of type singlet or singlets.
        Any singlets object with parent attribute as `None` is the topmost singlets object in the collection tree, else
        has some other singlets object as it's parent.

        The `reference` parameter specifies the argument pool from which the arguments are parsed, by default set to `sys.argv`
        with `skip` set to `1` which strips the `sys.argv[0]` value (script name). The `reference` parameter can be set to any
        iterable of strings with `skip` parameter to specify how many values from left side to skip.

        The `parent` must be another `singlets` object or simply `None`.
        
        Each `singlets` object contains a base argument which uniquely identifies the `singlets` collection. If the base argument
        is marked as `required=True` then it simply acts as a required argument. However, if any other argument in the collection
        is marked `required=True` then it is only required to pass that argument if the base argument is present (meaning: If the
        collection is being used (identified by usage of base argument), then only any required arguments are required to be passed).
        
        This passively (without using `required` parameter) applies to all nested `singlets` objects such that, any child `singlets`
        object cannot be used until it's parent's base argument has been used.
        """


    def __singlet__(self, name: str, *notations: str, base=False,
                    capture: Literal['auto', 'single', 'boolean'] = 'auto',
                    helptext: str = ..., required=False, default: Any = ...) -> Self:
        """Add a singlet object into the collection pool."""


    def __singlets__(self, name: str, parent: Union[singlets, None] = ..., helptext: str = ...) -> Self:
        """Add a child singlets object into the collection pool.
        
        If `parent` parameter is `None`, current `singlets` object will be set as the parent.
        """


    @overload
    def __helptext__(self, exit: Literal[True], exit_code=0, acknowledgement=True, usage=True) -> NoReturn:
        """Print helptext for the entire collection pool and exit."""

    @overload
    def __helptext__(self, exit: Literal[False], exit_code=0, acknowledgement=True, usage=True) -> None:
        """Print helptext for the entire collection pool."""


    @overload
    def __parse__(self, reference: Iterable[Union[singlet, singlets]]) -> Dict[str, Union[Dict[str, Any], Any]]:
        """Parse the entire collection pool and return a dict with names of objects as keys and their respective
        parsed values as keys. The `reference` parameter expects an iterable of all defined `singlet` and `singlets`
        object in the context."""

    @overload
    def __parse__(self, reference: Iterable[Union[singlet, singlets]], order: Iterable[Union[str, Tuple[Union[Tuple[Any, ...], str], ...]]]) -> Tuple:
        """Parse the entries from collection pool in given order where the `order` parameter accepts,
        object names. Returns a tuple with parsed values in sequence.
        
        This method is a bit complex but is very powerful in terms of retreiving parsed values of highly
        nested `singlets` objects.

        As an example, if you have 3 arguments named arg1, arg2 and arg3 and another `singlets` object as
        a child with name, say, child_singlet. Now that child_singlet further has arg4 and arg5.

        To get the parsed values in order of -> arg1, arg2, arg3, arg4, arg5

        ```python
        # you can either use:
        obj.__parse__(reference, order=('arg1', 'arg2', 'arg3', 'arg4', 'arg5'))
        # will return (v1, v2, v3, v4, v5)

        # or
        obj.__parse__(reference, order=('arg1', 'arg2', 'arg3', ('child_singlet', 'arg4', 'arg5')))
        # will return (v1, v2, v3 (v4, v5)).
        ```

        and so on.
        """


    def __object__(self, name: str, /) -> Union[singlet, singlets, None]:
        """`singlets` class' native object search function that returns the object with given name if found, else `None`."""


    def __register_singlet_for_this_singlets_object__(self, singlets_object_name: str, singlet_object_name: str, *notations: str,
                                                      base=False, capture: Literal['auto', 'single', 'boolean'] = 'auto', helptext: str = ..., default: Any = ...) -> Self:
        """Add a `singlet` object under given `singlets` object. If the `singlets` object name provided is of the current object then current object's `__singlet__`
        method will be called instead."""


    def __register_singlets_for_this_singlets_object__(self, parent_singlets_object_name, child_singlets_object_name, helptext: str = ...) -> Self:
        """Add a `singlets` object under given parent `singlets` object. If the parent `singlets` object name is of the current object then current object's
        `__singlets__` method will be called."""


import typing
import shutil
import sys

from lawyer.exceptions import ParameterError, ArgumentError, ParseError

# ----------------------------LICENSE---------------------------------------------

# MIT License

# Copyright (c) 2025 Soumyo Deep Gupta

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# ------------------------------END-----------------------------------------------

default = typing.TypeVar('default')


class singlet(typing.Generic[default]):


    # These attributes are available to access using sqare brackets on any
    # instance of `singlet` class.


    __attrs__ = ('name', 'notations', 'capture-type', 'helptext', 'default', 'required', 'reference', 'auto-help', 'help-notations',
                 'application-name', 'application-version', 'application-description', 'present', 'notation', 'notpresent', 'index')


    # These values are the only allowed values that the `capture` parameter
    # can take


    __allowed_capture_types__ = ('auto', 'single', 'boolean')


    # This is a special `__getitem__` method which facilitates accessing attributes
    # or custom named properties using square brackets.
    # All attributes in __attrs__ are allowed to be fetched.


    def __getitem__(self, name, /):
        if name not in self.__attrs__:
            raise AttributeError(
                f"'{name}' attribute cannot be accessed using square brackets.")
        if name is 'present':
            return self.__present__
        if name is 'notpresent':
            return self.__notpresent__
        if name is 'notation':
            return self.__notation__
        if name is 'index':
            return self.__index__
        return getattr(self, name.replace('-', '_'), None)


    # This is a special __setitem__ method which facilitates assignment of attributes
    # and custom named properties using square brackets.
    # Only some of the properties and attributes can be modified such as
    # ('present', 'notpresent', 'notation', 'index') cannot be modified.
    # After every change the __partial__ initialization method is called to retrieve
    # new information about the argument from self.reference.


    def __setitem__(self, name, value, /) -> None:
        if name in ('present', 'notpresent', 'notation', 'index') or name not in self.__attrs__:
            raise AttributeError(
                f"'{name}' attribute cannot be set using square brackets.")
        if name is 'name' and not isinstance(value, str):
            raise AttributeError(f"name attribute must be of type <class 'str'>.")
        if name is 'notations' and not (isinstance(value, typing.Iterable) and not isinstance(value, str) and all(isinstance(notation, str) for notation in value)):
            raise AttributeError(f"notations attribute must be an iterable of strings.")
        if name is 'capture-type' and value not in ['auto', 'single', 'boolean']:
            raise AttributeError(f"capture-type must be any one of 'auto', 'single' or 'boolean'.")
        if name is 'helptext' and not isinstance(value, str):
            raise AttributeError(f"helptext attribute must be of type <class 'str'>.")
        if name is 'required' and not isinstance(value, bool):
            raise AttributeError(f"required attribute must of type <class 'bool'>.")
        if name is 'reference' and not (isinstance(value, typing.Iterable) and not isinstance(value, str) and all(isinstance(ref, str) for ref in value)):
            raise AttributeError(f"reference attribute must be an iterable of strings.")
        if name is 'auto-help' and not isinstance(value, bool):
            raise AttributeError(f"auto-help attribute must be of type <class 'bool'>.")
        if name is 'help-notations' and not (isinstance(value, typing.Iterable) and not isinstance(value, str) and all(isinstance(help_notation, str) for help_notation in value)):
            raise AttributeError(f"help-notations must be an iterable of strings.")
        if name is 'application-name' and not isinstance(value, str):
            raise AttributeError(f"application-name attribute must be of type <class 'str'>.")
        if name is 'application-version' and not self.__version_format_checker__(value):
            raise AttributeError(f"application-version attribute must be of type <class 'str'> and in valid version format.")
        if name is 'application-description' and not isinstance(value, str):
            raise AttributeError(f"application-description attribute must be of type <class 'str'>.")
        setattr(self, name.replace('-', '_'), value)
        self.__partial__(raise_error=True)


    # No attribute or custom named properties will be allowed to be deleted
    # using square brackets.


    def __delitem__(self, name, /) -> None:
        raise AttributeError(
            f"attributes cannot be deleted using square brackets.")


    # The str function will return `singlet(*notations)`


    def __str__(self) -> str:
        return 'singlet(' + ', '.join(self.notations) + ')'


    # The string representation will return what str function will return when
    # used on self.


    def __repr__(self) -> str:
        return str(self)


    # The __readonly__ property construct is designed to be immutable and its core job is
    # to raise an AttributeError upon any kind of access/ assignment/ deletion.
    # This property will be used by other properties to generate exception when assignment
    # or deletion is being performed.


    @property
    def __readonly__(self) -> typing.NoReturn:
        raise AttributeError(f"No attribute named __readonly__ found for singlet class.")

    @__readonly__.setter
    def __readonly__(self, value) -> typing.NoReturn:
        raise AttributeError(f"properties of singlet class are read-only and doesn't support assignment.")

    @__readonly__.deleter
    def __readonly__(self) -> typing.NoReturn:
        raise AttributeError(f"properties of singlet class are read-only and doesn't support deletion.")


    # This method checks if any given object is a valid version string such as '1.0', '1.2.3',
    # and is currently half assed, will be removed further.


    @staticmethod
    def __version_format_checker__(candidate) -> typing.TypeGuard[str]:
        if not isinstance(candidate, str):
            return False
        if not '.' in candidate:
            return False
        # later on add more checks here.
        return True


    # The __init__ method will check all parameters and and store them. Additionally,
    # it will perform partial parsing where it will identify the presence status,
    # index and notation that is present in the reference.


    def __init__(self, name, *notations, capture='auto', helptext='', required=False, default: default = ..., reference=sys.argv,
                 skip=1, auto_help=False, help_notations=(), application_name=sys.argv[0], application_version='1.0', application_description='') -> None:
        if not isinstance(name, str):
            raise ParameterError(ParameterError.__strerror__.format('name'))
        if not all(isinstance(notation, str) for notation in notations):
            raise ParameterError(
                ParameterError.__strerror__.format('notations'))
        if not capture in self.__allowed_capture_types__:
            raise ParameterError(
                ParameterError.__capturetyperror__.format(
                    self.__allowed_capture_types__))
        if not isinstance(helptext, str):
            raise ParameterError(
                ParameterError.__strerror__.format('help-text'))
        if not isinstance(required, bool):
            raise ParameterError(
                ParameterError.__boolerror__.format('required parameter'))
        if not isinstance(reference, typing.Iterable) or not all(
                isinstance(ref, str) for ref in reference):
            raise ParameterError(
                ParameterError.__iterablerror__.format('reference'))
        if not isinstance(skip, int):
            raise ParameterError(
                'skip parameter must be of type <class \'int\'>.')
        if not isinstance(auto_help, bool):
            raise ParameterError(
                'auto-help parameter must be of type <class \'bool\'>.')
        if not isinstance(help_notations, typing.Iterable) or not all(
                isinstance(help_notation, str) for help_notation in help_notations):
            raise ParameterError(
                ParameterError.__iterablerror__.format('help-notations'))
        if not isinstance(application_name, str):
            raise ParameterError(
                ParameterError.__strerror__.format('application-name'))
        if not isinstance(application_version, str) or not self.__version_format_checker__(
                application_version):
            raise ParameterError(
                "application-version provided is not in valid version format such as 1.0, 2.8.9, etc.")
        if not isinstance(application_description, str):
            raise ParameterError(
                ParameterError.__strerror__.format('application-description'))

        self.name = name
        self.notations = notations
        self.capture_type = capture
        self.helptext = helptext
        
        if default is ... and self.capture_type is 'boolean':
            self.default = False
        elif default is not ... and self.capture_type is 'boolean':
            self.default = default

        self.required = required
        self.reference = reference[skip:]
        self.auto_help = auto_help
        self.help_notations = help_notations or ('--help', '-h')
        self.application_name = application_name
        self.application_version = application_version
        self.application_description = application_description
        self.__partial__(raise_error=True)


    # The __helptext__ method is responsible for printing current singlet object's
    # help text.


    def __helptext__(self, exit, exit_code=0) -> None:
        print(f"\n{self.application_name} v{self.application_version}")
        print(f"{self.application_description}\n")

        adder = ''.join(
            f"[{notation}]" for notation in self.help_notations) if self.help_notations else ''

        print(f"usage: {self.application_name} {self.__usage__}{adder}\n")

        helptext = ' | '.join(
            self.help_notations) if self.help_notations else ''

        selftext = ' | '.join(self.notations)
        maxlength = max(len(helptext), len(selftext), 0)

        if helptext:
            print(f"{helptext.ljust(maxlength + 3)} : show this helptext and exit.")
        print(f"{selftext.ljust(maxlength + 3)} : {self.helptext}")

        if exit:
            sys.exit(exit_code)


    # The __present__ property returns True if the singlet argument is found in the
    # reference else False


    @property
    def __present__(self) -> bool:
        return self.__ispresent__

    @__present__.setter
    def __present__(self, value) -> typing.NoReturn:
        self.__readonly__ = value

    @__present__.deleter
    def __present__(self) -> typing.NoReturn:
        del self.__readonly__


    # The __notation property returns the notation variant found in the reference
    # for this singlet argument, If not found returns an empty string


    @property
    def __notation__(self) -> str:
        if hasattr(self, '__notation_in_ref__'):
            return self.__notation_in_ref__
        return ''

    @__notation__.setter
    def __notation__(self, value) -> typing.NoReturn:
        self.__readonly__ = value

    @__notation__.deleter
    def __notation__(self) -> typing.NoReturn:
        del self.__readonly__


    # The __index__ property returns the index of reference where the singlet
    # object's one of the notation has been found. If not found, returns -1.


    @property
    def __index__(self) -> int:
        if hasattr(self, '__index_in_ref__'):
            return self.__index_in_ref__
        return -1

    @__index__.setter
    def __index__(self, value) -> typing.NoReturn:
        self.__readonly__ = value

    @__index__.setter
    def __index__(self) -> typing.NoReturn:
        del self.__readonly__


    # The __partial__ method is responsible for extracting information such as
    # the index, notation and presence status of the argument in the given reference
    # and is triggered as early as __init__. Updation of class attributes using square
    # brackets also triggers it.


    def __partial__(self, raise_error=True) -> None:

        if hasattr(self, '__notation_in_ref__'):
            del self.__notation_in_ref__
        if hasattr(self, '__index_in_ref__'):
            del self.__index_in_ref__

        if self.capture_type is 'boolean' and not hasattr(self, 'default'):
            self.default = False

        self.__ispresent__ = False
        for notation in self.notations:
            if notation in self.reference:
                self.__ispresent__ = True
                self.__notation_in_ref__ = notation
                self.__index_in_ref__ = self.reference.index(notation)
                break

        if self.__present__ and self.auto_help and self.help_notations:
            possible_help_trigger = self.__parse__(
                self, capture_type='single', raise_errors=False)
            if possible_help_trigger and possible_help_trigger in self.help_notations:
                self.__helptext__(exit=True, exit_code=0)

        if self.required and not self.__present__ and not hasattr(
                self, 'default') and raise_error:
            raise ArgumentError(
                f"Any one of {
                    self.notations} is a mandatory argument in this context!")


    # The __usage__ property returns only the usage of this particular argument in string
    # format.


    @property
    def __usage__(self) -> str:

        string = '[ '
        if len(self.notations) == 1:
            if self.capture_type is 'boolean':
                return string + self.notations[0] + ' ]'
            elif self.capture_type is 'single':
                return string + self.notations[0] + ' [VALUE] ]'
            else:
                return string + self.notations[0] + ' [...] ]'

        for notation in self.notations:
            if self.capture_type is 'boolean':
                string += f"[{notation}]"
            elif self.capture_type is 'single':
                string += f'[{notation} [VALUE]]'
            else:
                string += f'[{notation} [...]]'
        return string + ' ]'

    @__usage__.setter
    def __usage__(self, value) -> typing.NoReturn:
        self.__readonly__ = value

    @__usage__.deleter
    def __usage__(self) -> typing.NoReturn:
        del self.__readonly__


    # A function that checks if a given value is present in any of notations of any of the singlet/ singlets
    # objects in provided reference.


    @staticmethod
    def __value_in_reference__(*reference, value) -> bool:
        for ref in reference:
            if isinstance(ref, singlet) and ref.__present__ and value in ref.notations:
                return True
            elif isinstance(ref, singlets):
                is_true = singlet.__value_in_reference__(
                    *ref.collection.values(), value=value)
                if is_true:
                    return True
        return False


    # A function that returns a tuple of indices from self.reference for the known singlet/ singlets
    # notations where their index > index_to_check.
    # Meaning: return all indices for singlet/ singlets objects (that are provided in the reference
    # parameter) such that they are positionally on the right of the index_to_check parameter.


    @staticmethod
    def __indices_of_reference__(
            *reference, index_to_check) -> typing.Tuple[int, ...]:
        # will return indices of objects greater than index_to_check
        indices = []
        for ref in reference:
            if isinstance(
                    ref, singlet) and ref.__present__ and ref.__index__ > index_to_check:
                indices.append(ref.__index__)
            elif isinstance(ref, singlets):
                tree_indices = singlet.__indices_of_reference__(
                    *ref.collection.values(), index_to_check=index_to_check)
                for idx in tree_indices:
                    if idx not in indices:
                        indices.append(idx)
        return tuple(indices)


    # The __parse__ method is responsible for fully parsing the current singlet argument based on relative
    # reference and return the result.


    def __parse__(self, *reference, capture_type=None,
                  raise_errors=True) -> typing.Union[bool, str, typing.Tuple[str, ...], default, None]:
        if capture_type is None:
            capture_type = self.capture_type

        if self.__present__:

            if capture_type is 'boolean':
                return self.__present__

            elif capture_type is 'single':

                if not reference:
                    if raise_errors:
                        raise ParseError(
                            f"a reference is required for parsing capture-type='single'.")
                    else:
                        return None

                try:
                    potential_value = self.reference[self.__index__ + 1]
                    if self.__value_in_reference__(
                            *reference, value=potential_value):
                        if raise_errors:
                            raise ParseError(
                                f"'{self.__notation__}' expects a value.")
                        else:
                            return None
                    return potential_value
                except IndexError:
                    if raise_errors:
                        raise ParseError(
                            f"'{self.__notation__}' expects a value.")
                    else:
                        return None

            elif capture_type is 'auto':
                if not reference:
                    if raise_errors:
                        raise ParseError(
                            f"a reference is required for parsing capture-type='auto'.")
                values = self.reference[self.__index__ +
                                        1: min(self.__indices_of_reference__(*
                                                                             reference, self.__index__), default=len(self.reference))]
                if values:
                    if isinstance(values, typing.Iterable) and not isinstance(
                            values, str):
                        if len(values) > 1:
                            return tuple(values)
                        elif len(values) == 1:
                            return values[0]
                        else:
                            raise ParseError(
                                f"'{self.__notation__}' expects a value.")
                    else:
                        return values
                raise ParseError(
                    f"'{self.__notation__}' expects atleast one value.")

        elif hasattr(self, 'default'):
            return self.default

        return None


    # The __notpresent__ property is a compliment of __present__ property.


    @property
    def __notpresent__(self) -> bool:
        return not self.__present__

    @__notpresent__.setter
    def __notpresent__(self, value) -> typing.NoReturn:
        self.__readonly__ = value

    @__notpresent__.deleter
    def __notpresent__(self) -> typing.NoReturn:
        del self.__readonly__





class singlets:


    # All attributes that can be accessed via square brackets can be found in __attrs__.


    __attrs__ = ('name', 'parent', 'helptext', 'reference', 'auto-help', 'help-notations', 'application-name',
                 'application-version', 'application-description', 'collection', 'present', 'notpresent', 'base')


    # The __getitem__ method is responsible for providing access to the class attributes
    # using square brackets.


    def __getitem__(self, name, /):
        if name not in self.__attrs__:
            raise AttributeError(f"attribute {name} cannot be accessed using sqare brackets.")
        if name is 'present':
            return self.__ispresent__
        if name is 'notpresent':
            return not self.__ispresent__
        if name is 'base':
            return self.__base__

        return getattr(self, name.replace('-', '_'), None)


    # The __setitem__ method is responsible for enforcing assignment rules for class attributes
    # via square brackets.


    def __setitem__(self, name, value, /) -> None:
        if name in ('present', 'notpresent', 'collection') or name not in self.__attrs__:
            raise AttributeError(f"attribute {name} cannot be set using sqare brackets.")
        if name is 'name' and not isinstance(value, str):
            raise AttributeError(f"name of the singlets collection must be of type <class 'str'>.")
        if name is 'parent' and not (isinstance(value, singlets) or value is None):
            raise AttributeError(f"parent of a singlets object can only be another singlets object, or None.")
        if name is 'helptext' and not isinstance(value, str):
            raise AttributeError(f"help-text of singlets object must be of type <class 'str'>.")
        if name is 'reference' and not (isinstance(value, typing.Iterable) and all(isinstance(ref, str) for ref in value)):
            raise AttributeError(f"reference attribute of singlets object must be an iterable of strings.")
        if name is 'auto-help' and not isinstance(value, bool):
            raise AttributeError(f"auto-help attribute expects a bool value.")
        if name is 'help-notations' and not (isinstance(value, typing.Iterable) and all(isinstance(help_notation, str) for help_notation in value)):
            raise AttributeError(f"help-notations attribute expects an iterable of strings.")
        if name is 'application-name' and not isinstance(value, str):
            raise AttributeError(f"application-name attribute expects a string value.")
        if name is 'application-version' and not singlet.__version_format_checker__(value):
            raise AttributeError(f"application-version expects a string value which is in valid version format.")
        if name is 'application-description' and not isinstance(value, str):
            raise AttributeError(f"application-description attribute expects a string value.")
        if name is 'base' and not isinstance(value, singlet):
            raise AttributeError(f"base attribute expects a singlet object.")

        if name is 'base':
            if self.base is not None:
                del self.collection[self.base]
            self.base = value.name
            self.collection[self.base] = value
            return None

        setattr(self, name.replace('-', '_'), value)


    # The __delitem__ method prevents any deletion of attributes


    def __delitem__(self, name, /) -> typing.NoReturn:
        raise AttributeError(f"cannot delete attribute {name} of singlets class.")


    # The __readonly__ property is a construcy that restricts all kinds of modification
    # to any given property making it readonly.


    @property
    def __readonly__(self) -> typing.NoReturn:
        raise AttributeError(f"No property named __readonly__ found in singlets class.")

    @__readonly__.setter
    def __readonly__(self, value) -> typing.NoReturn:
        raise AttributeError(f"properties of singlets class are read-only and doesn't support assignment.")

    @__readonly__.deleter
    def __readonly__(self) -> typing.NoReturn:
        raise AttributeError(f"properties of singlets class are read-only and doesn't suport deletion.")


    # The __base__ property returns the base argument object (singlet) if defined, else
    # None.


    @property
    def __base__(self) -> typing.Union[singlet, None]:
        if self.base is None:
            return None
        return self.collection[self.base]

    @__base__.setter
    def __base__(self, value) -> typing.NoReturn:
        self.__readonly__ = value

    @__base__.deleter
    def __base__(self) -> typing.NoReturn:
        del self.__readonly__


    # The __init__ method in singlets' context only checks the parameters and stores them.

    def __init__(self, name, parent=None, helptext='', reference=sys.argv, skip=1, auto_help=False,
                 help_notations=(), application_name=sys.argv[0], application_version='1.0', application_description='') -> None:

        if not isinstance(name, str):
            raise ParameterError(ParameterError.__strerror__.format('name'))
        if not isinstance(helptext, str):
            raise ParameterError(
                ParameterError.__strerror__.format('help-text'))
        if not isinstance(reference, typing.Iterable) or not all(
                isinstance(ref, str) for ref in reference):
            raise ParameterError(
                ParameterError.__iterablerror__.format('reference'))
        if not isinstance(skip, int):
            raise ParameterError(
                'skip parameter must be of type <class \'int\'>.')
        if not isinstance(auto_help, bool):
            raise ParameterError(
                'auto-help parameter must be of type <class \'bool\'>.')
        if not isinstance(help_notations, typing.Iterable) or not all(
                isinstance(help_notation, str) for help_notation in help_notations):
            raise ParameterError(
                ParameterError.__iterablerror__.format('help-notations'))
        if not isinstance(application_name, str):
            raise ParameterError(
                ParameterError.__strerror__.format('application-name'))
        if not isinstance(application_version, str) or not singlet.__version_format_checker__(
                application_version):
            raise ParameterError(
                "application-version provided is not in valid version format such as 1.0, 2.8.9, etc.")
        if not isinstance(application_description, str):
            raise ParameterError(
                ParameterError.__strerror__.format('application-description'))
        
        if not isinstance(parent, singlets) and parent is not None:
            raise ParameterError(f"parent parameter must be of the type 'singlets' or None.")

        self.name = name
        self.parent = parent
        self.helptext = helptext
        self.reference = reference[skip:]

        self.collection: typing.Dict[str, typing.Union[singlet, singlets]] = {}
        self.base = None

        self.__ispresent__ = False
        self.auto_help = auto_help
        self.help_notations = help_notations
        self.application_name = application_name
        self.application_version = application_version
        self.application_description = application_description


    # The __singlet__ method is responsible for registering a singlet object into the
    # collection.


    def __singlet__(self, name, *notations, base=False, capture='auto',
                    helptext='', required=False, default=...) -> typing.Self:
        if not isinstance(base, bool):
            raise ParameterError(
                '\'base\' parameter must be of type <class \'bool\'>.')

        if base:
            if self.base is None:
                self.base = name
            else:
                raise RuntimeError(f"base argument for group '{self.name}' has already been registered.")

        self.collection[name] = singlet(
            name,
            *notations,
            capture=capture,
            helptext=helptext,
            required=required if base else False,
            default=default,
            reference=self.reference,
            skip=0,
            auto_help=False,
            help_notations=self.help_notations,
            application_name=self.application_name,
            application_version=self.application_version,
            application_description=self.application_description)
        
        if self.base is name:
            self.__ispresent__ = self.collection[self.base].__ispresent__


        # check for requirement
        # only for other arguments than the base, as it will be handled
        # by singlet itself.
        if not base and required and self.__present__ and self.collection[name].__notpresent__:
            raise ArgumentError(f"Any one of {notations} is mandatory in when using {self.collection[self.base].__notation__}.")

        return self


    # The __singlets__ method is responsible for registeing a singlets object into
    # the collection.


    def __singlets__(self, name, parent=None, helptext='') -> typing.Self:
        if parent is None:
            parent = self

        self.collection[name] = singlets(name, parent, helptext, self.reference, 0, self.auto_help,
                                         self.help_notations, self.application_name, self.application_version, self.application_description)
        return self


    # The __helptext__ method is responsible for printing the collective helptext of all objects
    # in the collection and exit.


    def __helptext__(self, exit, exit_code=0, acknowledgement=True, usage=True) -> None:
        if self.base is None:
            raise ArgumentError(f"No base argument defined for argument group '{self.name}'; Failure to generate heltext.")

        if acknowledgement:
            print(f"\n{self.application_name} v{self.application_version}")
            print(f"{self.application_description}\n")


        adder = ''.join(f"[{notation}]" for notation in self.help_notations) if self.help_notations else ''

        usage_string_list = []
        terminal_width = shutil.get_terminal_size().columns - 10 - len(self.application_name) - 1

        usage_text = ''
        notation_texts = [(f" | ".join(self.help_notations), "show this helptext and exit.")] if self.help_notations else []
        maxlength = max((len(n) for n, _ in notation_texts), default=0)

        groups: typing.List[singlets] = []

        # show info about the parent if present, that this group
        # can only be used with the parent group.

        usage_text += self.collection[self.base].__usage__
        
        for arg_or_grp in self.collection.values():
            if isinstance(arg_or_grp, singlet):
                if arg_or_grp.name is not self.base:
                    _local_usage_text = arg_or_grp.__usage__
                    if len(usage_text + _local_usage_text) > terminal_width:
                        usage_string_list.append(usage_text)
                        usage_text = _local_usage_text
                    else:
                        usage_text += _local_usage_text
                
                notation = ' | '.join(arg_or_grp.notations)
                notation_texts.append((notation, arg_or_grp.helptext))
                maxlength = max(maxlength, len(notation))

            else:
                groups.append(arg_or_grp)


        if len(usage_text + adder) > terminal_width:
            usage_string_list.extend([usage_text, adder])
        else:
            usage_string_list.append(usage_text + adder)

        if usage:
            print(f"usage:".ljust(9), self.application_name, usage_string_list[0])
            for line in usage_string_list[1:]:
                print(''.ljust(10 + len(self.application_name)), line)

        if self.parent is not None:
            print("\n" if usage else "", 'syntax: ...', '<' + self.parent.name + ' group (parent)>', '<' + self.name + ' group>')
            print("This argument group can only be used when the parent group is used.")

        print(f"\n{self.name}\n{self.helptext}\n")

        for notation, text in notation_texts:
            print(f"{notation.ljust(maxlength + 3)}: {text}")


        if groups:
            for grp in groups:
                grp.__helptext__(exit=False, acknowledgement=False, usage=False)

        if exit is True:
            sys.exit(exit_code)


    # The __parse__ method is responsible for handling full fledged parsing of all objects present
    # in the collection and return it.


    def __parse__(self, reference, order=None):
        # Format:
        # order = ( A, B, (C, a, b), D)
        # where A, B, D are singlet and (C, a, b) is a singlets object with C as its name.
        # Returns: (rA, rB, rC, rD), where r<anything> is result.


        if order:
            values = []
            for order_item in order:
                
                if isinstance(order_item, tuple):
                    name = order_item[0]
                    child_order = order_item[1:]

                    recursive_object = self.__object__(name)

                    # must exist
                    if recursive_object is None:
                        raise ParseError(f"no object with name {name} is registered in the collection pool.")
                    
                    # must be singlets
                    if not isinstance(recursive_object, singlets):
                        raise ParseError(f"object with name {name} is not a singlets object, a tuple structure is not necessary.")
                    
                    values.append(recursive_object.__parse__(reference, order=child_order))
                    continue

                obj = self.__object__(order_item)

                if obj is None:
                    raise ParseError(f"no object with name {name} is registered in the collection pool.")

                if isinstance(obj, singlet):

                    if obj.__present__ and self.auto_help and self.help_notations and obj.__parse__(*reference, capture_type='single', raise_errors=False) in self.help_notations:
                        self.__helptext__(exit=True)

                    values.append(obj.__parse__(*reference))
                else:
                    raise ParseError(f"singlets object with name {name} cannot be parsed fully as order is being used. Instead of providing just the name of the singlets object, provide a tuple, where the first element is the singlet's object name and all other elements represent the order of that particular singlets object.")
            return tuple(values)
        else:
            values = {}
            for name, obj in self.collection.items():
                if isinstance(obj, singlet):
                    if obj.__present__ and self.auto_help and self.help_notations and obj.__parse__(*reference, capture_type='single', raise_errors=False) in self.help_notations:
                        self.__helptext__(exit=True)

                    values[name] = obj.__parse__(*reference)
                else:
                    values[name] = obj.__parse__(reference)
            return values


    # The __object__ method is responsible for returning the object with given name,
    # no matter where and how deep it exists starting its search from the collection.


    def __object__(self, name, /) -> typing.Union[singlet, 'singlets', None]:
        for _, obj in self.collection.items():
            if isinstance(obj, singlet) and obj.name is name:
                return obj
            elif isinstance(obj, singlets):
                if obj.name is name:
                    return obj
                
                potential_object = obj.__object__(name)
                if potential_object:
                    return potential_object
        return None


    # The below method is responsible for registering a singlet object under
    # a given singlets object (the singlets object can be present anywhere under
    # the collection tree).


    def __register_singlet_for_this_singlets_object__(self, singlets_object_name, singlet_object_name, *notations,
                                                      base=False, capture='auto', helptext='', required=False, default=...) -> typing.Self:
        if singlets_object_name is self.name:
            return self.__singlet__(singlet_object_name, *notations, base=base, capture=capture,
                                    helptext=helptext, required=required, default=default)
        
        potential_object = self.__object__(singlets_object_name)
        if potential_object and isinstance(potential_object, singlets):
            potential_object.__singlet__(singlet_object_name, *notations, base=base, capture=capture, helptext=helptext, required=required, default=default)
            return self
        raise RuntimeError(f"Either singlets object with name {singlets_object_name} does not exist or is not a singlets object at all.")


    # The below method is responsible for registering a singlets object under
    # a given singlets object (the parent singlets object can be present anywhere
    # under the collection tree).


    def __register_singlets_for_this_singlets_object__(self, parent_singlets_object_name, child_singlets_object_name, helptext='') -> typing.Self:
        if parent_singlets_object_name is self.base:
            return self.__singlets__(child_singlets_object_name, self, helptext=helptext)
        
        potential_object = self.__object__(parent_singlets_object_name)
        if potential_object and isinstance(potential_object, singlets):
            potential_object.__singlets__(child_singlets_object_name, potential_object, helptext)
            return self
        raise RuntimeError(f"Either singlets object with name {parent_singlets_object_name} does not exist or is not a singlets object at all.")




from lawyer.parsers import singlet, singlets
from lawyer.exceptions import InstanceError, ParameterError, ExistsError, ArgumentError, ParseError
import sys, typing, traceback, linecache


_modify = False
original_excepthook = sys.excepthook

def minimal_traceback(exception: typing.Type[Exception], *args: object) -> typing.NoReturn:
    
    def internal_callback(exc_type, exc_value, exc_tb):

        # Get full traceback frames
        tb_list = traceback.extract_tb(exc_tb)

        # Get the outermost user-call frame (first frame in the traceback)
        user_frame = tb_list[0]  # index 0 = where error was triggered in user code

        # Print traceback header
        print("Traceback (most recent call last):")

        # Show only user's line (not internal frames)
        if sys.platform in ('darwin', 'linux'):
            import colorama
            colorama.init()
            print(f'  File "{colorama.Fore.RED}{user_frame.filename}{colorama.Fore.RESET}", line {colorama.Fore.RED}{user_frame.lineno}{colorama.Fore.RESET}, in {colorama.Fore.RED}{user_frame.name}{colorama.Fore.RESET}')
        else:
            print(f'  File "{user_frame.filename}", line {user_frame.lineno}, in {user_frame.name}')

        code_line = user_frame.line or linecache.getline(user_frame.filename, user_frame.lineno).strip()
        print(f'    {code_line}')
        if sys.platform in ('darwin', 'linux'):
            print('    ' + colorama.Fore.RED + ''.join('^' if c != ' ' else ' ' for c in code_line) + colorama.Fore.RESET)
        else:
            print('    ' + ''.join('^' if c != ' ' else ' ' for c in code_line))

        # Print final exception message
        if sys.platform in ('darwin', 'linux'):
            if exc_value:
                print(f"{colorama.Fore.RED}{exc_type.__name__}{colorama.Fore.RESET}: {colorama.Fore.RED}{exc_value}{colorama.Fore.RESET}")
            else:
                print(f"{colorama.Fore.RED}{exc_type.__name__}{colorama.Fore.RESET}")
            colorama.deinit()
        else:
            if exc_value:
                print(f"{exc_type.__name__}: {exc_value}")
            else:
                print(f"{exc_type.__name__}")

        global original_excepthool
        sys.excepthook = original_excepthook

    sys.excepthook = internal_callback
    raise exception(*args)


class Judge:


    @property
    def __readonly__(self) -> None:
        return None


    @__readonly__.setter
    def __readonly__(self, value, /) -> typing.NoReturn:
        return minimal_traceback(AttributeError, f"attributes of Judge class are read-only, cannot assign '{value}' to attribute.")


    @__readonly__.deleter
    def __readonly__(self) -> typing.NoReturn:
        return minimal_traceback(AttributeError, f"attributes of Judge class are read-only and frozen, cannot delete attribute.")


    @property
    def __application__(self) -> str:
        global _modify
        _modify = True
        value = getattr(self, '-8645880369160574959', sys.argv[0])
        _modify = False
        return value


    @__application__.setter
    def __application__(self, name, /) -> typing.NoReturn:
        if not isinstance(name, str):
            return minimal_traceback(AttributeError, f"application name must be of type <class 'str'>.")
        
        global _modify
        _modify = True
        setattr(self, '-8645880369160574959', name)
        _modify = False


    @__application__.deleter
    def __application__(self) -> typing.NoReturn:
        del self.__readonly__


    @property
    def __version__(self) -> str:
        global _modify
        _modify = True
        value = getattr(self, '-8059281209987537947', '1.0')
        _modify = False
        return value


    @__version__.setter
    def __version__(self, version, /) -> typing.NoReturn:
        if not singlet.__version_format_checker__(version):
            return minimal_traceback(AttributeError, f"version must be in valid format and must be of type <class 'str'>.")
        
        global _modify
        _modify = True
        setattr(self, '-8059281209987537947', version)
        _modify = False


    @__version__.deleter
    def __version__(self) -> typing.NoReturn:
        del self.__readonly__


    @property
    def __description__(self) -> str:
        global _modify
        _modify = True
        value = getattr(self, '-7681659577293593483', '')
        _modify = False
        return value


    @__description__.setter
    def __description__(self, desc, /) -> typing.NoReturn:
        if not isinstance(desc, str):
            return minimal_traceback(AttributeError, f"description must be of type <class 'str'>.")
        
        global _modify
        _modify = True
        setattr(self, '-7681659577293593483', desc)
        _modify = False


    @__description__.deleter
    def __description__(self) -> typing.NoReturn:
        del self.__readonly__


    @property
    def __reference__(self) -> typing.Iterable[str]:
        global _modify
        _modify = True
        value = getattr(self, '-4799753283422340294', sys.argv)
        _modify = False
        return value


    @__reference__.setter
    def __reference__(self, value, /) -> typing.NoReturn:
        self.__readonly__ = value


    @__reference__.deleter
    def __reference__(self) -> typing.NoReturn:
        del self.__readonly__


    def __setattr__(self, name: str, value: typing.Any, /) -> None:
        global _modify
        allowed_hashes = ('-7868989330756850304', '-8645880369160574959', '-8059281209987537947', '-7681659577293593483', '-4799753283422340294', '-1473320931206488208', '8587656167592366246')
        allowed_properties = ('__reference__', '__collection__', '__application__', '__version__', '__description__', '__readonly__', '__autohelp__', '__help_notations__')
        if name in allowed_hashes and _modify is False:
            return minimal_traceback(AttributeError, f"cannot assign '{value}' to a placeholder.")
        elif name not in allowed_hashes + allowed_properties:
            return minimal_traceback(AttributeError, f"cannot assign '{value}' to attribute {name} as Judge class is read-only.")
        return super().__setattr__(name, value)


    def __getattribute__(self, name: str, default=None, /) -> typing.Any:
        global _modify
        denied_hashes = ('-7868989330756850304', '-8645880369160574959', '-8059281209987537947', '-7681659577293593483', '-4799753283422340294', '-1473320931206488208', '8587656167592366246')
        if name in denied_hashes and _modify is False:
            return minimal_traceback(AttributeError, f"cannot access attribute {name}. object with name {name} is just a placeholder.")
        try:
            return super().__getattribute__(name)
        except Exception:
            return default


    @property
    def __collection__(self) -> typing.Dict[str, typing.Union[singlet, singlets]]:
        global _modify
        _modify = True
        if hasattr(self, '-7868989330756850304'):
            value = getattr(self, '-7868989330756850304')
            _modify = False
            return value
        else:
            setattr(self, '-7868989330756850304', {})
            value = getattr(self, '-7868989330756850304')
            _modify = False
            return value


    @__collection__.setter
    def __collection__(self, col, /) -> typing.NoReturn:
        self.__readonly__ = col


    @__collection__.deleter
    def __collection__(self) -> typing.NoReturn:
        del self.__readonly__


    @property
    def __autohelp__(self) -> bool:
        global _modify
        _modify = True
        if hasattr(self, '-1473320931206488208'):
            value = getattr(self, '-1473320931206488208')
        else:
            setattr(self, '-1473320931206488208', False)
            value = False
        _modify = False
        return value


    @__autohelp__.setter
    def __autohelp__(self, auto_help) -> None:
        if not isinstance(auto_help, bool):
            return minimal_traceback(AttributeError, f"__autohelp__ accepts boolean values only.")
        global _modify
        _modify = True
        setattr(self, '-1473320931206488208', auto_help)
        _modify = False


    @__autohelp__.deleter
    def __autohelp__(self) -> typing.NoReturn:
        del self.__readonly__


    @property
    def __help_notations__(self) -> typing.Iterable[str]:
        global _modify
        _modify = True

        if hasattr(self, '8587656167592366246'):
            value = getattr(self, '8587656167592366246')
        else:
            setattr(self, '8587656167592366246', ('--help', '-h'))
            value = ('--help', '-h')
        _modify = False
        return value


    @__help_notations__.setter
    def __help_notations__(self, help_notations) -> None:
        if not (isinstance(help_notations, typing.Iterable) and all(isinstance(help_notation, str) for help_notation in help_notations)):
            return minimal_traceback(AttributeError, f"__help_notations__ accepts an iterable of strings.")
        
        global _modify
        _modify = True
        setattr(self, '8587656167592366246', help_notations)
        _modify = False


    @__help_notations__.deleter
    def __help_notations__(self) -> typing.NoReturn:
        del self.__readonly__


    def __init__(self, application_name=sys.argv[0], application_version='1.0',
                 application_description='', reference=sys.argv, skip=1, auto_help=False, help_notations=()) -> None:

        if not isinstance(application_name, str):
            return minimal_traceback(ParameterError, f"application-name parameter must be of type <class 'str'>.")
        if not singlet.__version_format_checker__(application_version):
            return minimal_traceback(ParameterError, f"application-version parameter must a valid version string.")
        if not isinstance(application_description, str):
            return minimal_traceback(ParameterError, f"application-description parameter must be of type <class 'str'>.")
        if not (isinstance(reference, typing.Iterable) and all(isinstance(ref, str) for ref in reference)):
            return minimal_traceback(ParameterError, f"reference parameter must be an iterable of strings.")
        if not isinstance(skip, int):
            return minimal_traceback(ParameterError, "skip parameter must be of type <class 'int'>.")

        global _modify
        _modify = True

        setattr(self, '-8645880369160574959', application_name)
        setattr(self, '-8059281209987537947', application_version)
        setattr(self, '-7681659577293593483', application_description)
        setattr(self, '-4799753283422340294', reference[skip:])
        setattr(self, '-1473320931206488208', auto_help)
        setattr(self, '8587656167592366246', help_notations or ('--help', '-h'))
        setattr(self, '-7868989330756850304', {})


        _modify = False


    def add_argument(self, name: str, *notations: str, capture: typing.Literal['auto', 'single', 'boolean'] = 'boolean',
                     helptext='', required=False, default: typing.Any = ..., parent=None, base=False) -> typing.Self:
        
        if not (isinstance(parent, str) or parent is None):
            return minimal_traceback(ParameterError, f"parent can either be an argument group name or None.")
        
        if parent is not None:
            parent_object = self.object(parent)
            if parent_object is None:
                return minimal_traceback(InstanceError, f"object with name '{parent}' not found in the object tree.")
            if not isinstance(parent_object, singlets):
                return minimal_traceback(InstanceError, f"object with name '{parent}' is not an argument group.")
            
            if name in parent_object.collection:
                return minimal_traceback(ParameterError, f"argument or group with name '{name}' already exists under parent '{parent}'")
            
            try:
                parent_object.__singlet__(name, *notations, base=base, capture=capture, helptext=helptext, required=required, default=default)
            except ParameterError as p_error:
                return minimal_traceback(ParameterError, *p_error.args)
            except ArgumentError as a_error:
                return minimal_traceback(ArgumentError, *a_error.args)
            except RuntimeError as r_error:
                return minimal_traceback(RuntimeError, *r_error.args)
            except Exception as e:
                return minimal_traceback(Exception, *e.args)
            return self


        if name in self.__collection__:
            return minimal_traceback(ExistsError, f"argument or group with name '{name}' already exists.")

        try:
            self.__collection__[name] = singlet(name, *notations, capture=capture, helptext=helptext, required=required,
                                                default=default, auto_help=self.__autohelp__, help_notations=self.__help_notations__, application_name=self.__application__,
                                                application_version=self.__version__, application_description=self.__description__)
        except ParameterError as p_error:
            return minimal_traceback(ParameterError, *p_error.args)
        except ArgumentError as a_error:
            return minimal_traceback(ArgumentError, *a_error.args)
        except RuntimeError as r_error:
            return minimal_traceback(RuntimeError, *r_error.args)
        except Exception as e:
            return minimal_traceback(Exception, *e.args)
        return self

    def object(self, name: str, /) -> typing.Union[singlet, singlets, None]:
        if name in self.__collection__:
            return self.__collection__[name]
        
        for _object in self.__collection__.values():
            if isinstance(_object, singlets):
                potential_object = _object.__object__(name)
                if potential_object is not None:
                    return potential_object
        return None


    def add_group(self, name: str, parent=None, helptext='') -> typing.Self:
        if not (isinstance(parent, str) or parent is None):
            return minimal_traceback(ParameterError, f"parent can either be an argument group name or None.")
        
        if parent is not None:
            parent_object = self.object(parent)

            if parent_object is None:
                return minimal_traceback(InstanceError, f"object with name '{parent}' not found in the object tree.")
            if not isinstance(parent_object, singlets):
                return minimal_traceback(InstanceError, f"object with name '{parent}' is not an argument group.")
            
            if name in parent_object.collection:
                return minimal_traceback(ParameterError, f"argument or group with name '{name}' already exists under parent '{parent}'.")
            
            try:
                parent_object.__singlets__(name, helptext=helptext)
            except ParameterError as p_error:
                return minimal_traceback(ParameterError, *p_error.args)
            except Exception as e:
                return minimal_traceback(Exception, *e.args)
            return self


        if name in self.__collection__:
            return minimal_traceback(ParameterError, f"argument or group with name '{name}' already exists.")
        
        try:
            self.__collection__[name] = singlets(name, None, helptext, self.__reference__, 0, self.__autohelp__, self.__help_notations__, self.__application__, self.__version__, self.__description__)
        except ParameterError as p_error:
            return minimal_traceback(ParameterError, *p_error.args)
        except Exception as e:
            return minimal_traceback(Exception, *e.args)
        return self


    def parse(self, *order):
        if order:

            values = []

            for item in order:

                if isinstance(item, tuple) and all(isinstance(item_item, str) for item_item in item):
                    item_name = item[0]
                    item_order = item[1:]
                    potential_object = self.object(item_name)
                    if potential_object is None:
                        return minimal_traceback(ParseError, f"no object with name '{item_name}' found registered.")
                    elif isinstance(potential_object, singlet):
                        return minimal_traceback(ParseError, f"for parsing single arguments, their name is enough.")
                    else:
                        values.append(potential_object.__parse__(self.__collection__.values(), item_order))
                elif isinstance(item, str):
                    potential_object = self.object(item)
                    if potential_object is None:
                        return minimal_traceback(ParseError, f"no object with name '{item}' found registered.")
                    elif isinstance(potential_object, singlets):
                        return minimal_traceback(ParseError, f"for parsing grouped arguments use a tuple instead of just the name; The tuple's first element must be the group name and the item(s) to be parsed inside it should be after that.")
                    else:
                        values.append(potential_object.__parse__(*self.__collection__.values()))
                else:
                    return minimal_traceback(ParseError, f"the parse method accepts objects of type <class 'str'> or <class 'tuple'>.")

            if len(values) == 1:
                return values[0]
            else:
                return tuple(values)
        else:
            values = {}
            for item, item_object in self.__collection__.items():
                if isinstance(item_object, singlet):
                    values[item] = item_object.__parse__(*self.__collection__.values())
                else:
                    values[item] = item_object.__parse__(self.__collection__.values())
            return values



class ParameterError(Exception):

        __strerror__ = "{} for an argument or group must be of type <class 'str'>."
        __interror__ = __strerror__.replace('str', 'int')
        __boolerror__ = __strerror__.replace('str', 'bool')
        __iterablerror__ = __strerror__.replace("<class 'str'>", 'iterable[str]')
        __capturetyperror__ = "capture type of an argument must be any one of these: {}."

class ArgumentError(Exception): pass
class InstanceError(Exception): pass
class ParseError(Exception): pass
class ExistsError(Exception): pass


"""lawyer [`Module`].

This module contains the `Judge` class which is a safe implementation of collective
`lawyer.parsers.singlet` and `lawyer.parsers.singlets` classes. It also contains a
method named `minimal_traceback` which the `Judge` class uses for creating highly
focussed exception tracebacks to specify the exact high level line where error
occured.

----------------------------------------
This module is under license as follows:
----------------------------------------

MIT License
-----------

Copyright (c) 2025 Soumyo Deep Gupta

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
---------------------------- END of LICENSE ----------------------------------
"""



from lawyer.parsers import singlet, singlets
from typing import Iterable, Literal, Type, NoReturn, Dict
from typing import Union, overload, Any, Self, Tuple
import sys


original_excepthook = sys.excepthook

def minimal_traceback(exception: Type[BaseException], *args: object) -> NoReturn:
    """Raise the `exception` with given `args` limiting output to exactly last call of error."""


class Judge:
    """A safe implementation of argument + group parsing.
    
    This class covers a safe code implementation to avoid most errors and show the root cause
    of where the error actually occurred and why. Instead of maining a long definition of
    `singlet` and `singlets` objects to create argument parsing, this class acts as a wrapper
    class that handles most operations internally and provides high level user callable methods.

    `Judge` allows addition of any number of single or grouped arguments (which can be nested) and parsing
    in any order or individually. Most attributes and properties in this class are made read-only to prevent
    failure due to mishandling. Some properties allow modification such as `__application__` (application name),
    `__version__` (application version), `__description__` (application description), `__autohelp__` (whether to
    show help text automatically) and `__help_notations__` (the argument notation that denotes help).

    `Judge` also allows accessing internal `singlet` and `singlets` objects using the `object` method which requires
    the object's name.

    `Judge` class is not thread-safe. Creation of a single instance and using it across threads will result in
    malfuction.
    """

    @property
    def __application__(self) -> str:
        """The application name to be used during helptext generation. If not set, the default is
        set to `sys.argv[0]` (the script name)."""

    @property
    def __version__(self) -> str:
        """The application version to be used during helptext generation. If not set, the default is
        set to `1.0`."""

    @property
    def __description__(self) -> str:
        """The application description to be used during helptext generation. If not set, the default is
        empty and will not be shown."""

    @property
    def __reference__(self) -> Iterable[str]:
        """The argument pool from which arguments and groups are parsed. Once set using the `__init__`
        method, this property becomes read-only."""

    @property
    def __collection__(self) -> Dict[str, Union[singlet, singlets]]:
        """The internal argument/group collection tree; read-only."""

    @property
    def __autohelp__(self) -> bool:
        """The internal flag that determines whether to show helptext automatically. If set to `True`, requires
        to update `__help_notations__` property else will default to `--help` and `-h`."""

    @property
    def __help_notations__(self) -> Iterable[str]:
        """The notations to use for help argument detection. Defaults to `--help` and `-h`."""

    def __setattr__(self, name: str, value: Any, /) -> None: ...
    def __getattribute__(self, name: str, default: Any = None, /) -> Any: ...


    def __init__(self, application_name: str = ..., application_version: str = ...,
                 application_description: str = ..., reference: Iterable[str] = ...,
                 skip: int = ..., auto_help=False, help_notations: Iterable[str] = ...) -> None:
        """Create a `Judge` instance with given parameters. Some of these parameters become
        read-only upon object creation and therefore pay utmost attention to those.
        
        Attributes that becomes read-only:
        - `reference`: The argument pool from which argument(s)/group(s) will be parsed. By default it assumes
        the value of `sys.argv` with `skip` number of elements skipped from it from the left. By default `skip`
        assumes the value of `1` and therefore the resultant default argument pool is `sys.argv[1:]`. Can be
        accessed later using the `__reference__` property, but cannot be changed or deleted.

        Attributes the does not become read-only:
        - `application_name`: This is the application name to be used for helptext generation. If not set during
        `__init__`, can be set or modified later using the `__application__` property. If not changed throughout,
        will use `sys.argv[0]` as its default value.

        - `application_version`: This is the application version to be used for helptext generation. If not set
        during `__init__`, can be set or modified later using `__version__` property. If not changed throughout,
        will use `1.0` as its default value.

        - `application_description`: This description will be used during helptext generation if given as is totally
        optional. If not set during `__init__`, can be set or modified later using `__description__` property. If not
        changed throughout, will be empty and will not be used.

        - `auto_help`: Specifies whether to automatically show helptext and exit the script if any of the `help_notations`
        are found after an argument/group-argument. If set to `True`, and `help_notations` are not set, default help
        notations will be used: `('--help', '-h')`. Can be changed or set later using the `__autohelp__` property.

        - `help_notations`: Specifies a set of notations that do not collide with other argument/group definitions, that
        uniquely identifies if the user is asking for helptext for any given argument/group. By default, assumes notations:
        `--help` and `-h` for helptext, but wont be triggered until `auto_help` is set to `True`. Can be changed or set later
        using `__help_notations__` property.
        """

    def object(self, name: str, /) -> Union[singlet, singlets, None]:
        """Returns the underlying object with given name. If not found, returns `None`."""

    def add_argument(self, name: str, *notations: str, capture: Literal['auto', 'single', 'boolean'] = 'boolean',
                     helptext: str = ..., required=False, default: Any = ..., parent: str = ..., base=False) -> Self:
        """Add an argument into the collection. The parameters are very strict and cannot be changed through this
        class. However, the `object` method of this class can be used to retreive the internal underlying argument/
        group object and then parameters can be modified.
        
        Parameters:
        - `name`: name of the argument, make this unique as the argument will be accessed using this parameter only. Acts as
        an identifier for the argument.

        - `*notations`: argument notations to look for in the argument pool such as `--argument`. There is no constraint over
        how the argument notations needs to be and therefore simple words such as `argument` can also be used or any custom
        kind such as `@argument`. supports intermixed or multiple signatures.

        ```zsh
        $ your-script.py @argument1 --argumemt2 value # example.
        ```

        - `capture`: This parameter sets the argument capture mode. If the argument is supposed to take some value(s), use either
        `'single'` (expects exactly one value after the argument) or `'auto'` (expects atleast one value after the argument).
        Else if the argument needs to act as a flag such that it's presence in the argument pool represents boolean `True` and absence
        represents `False`, use `'boolean'` capture mode.

        - `helptext`: The help description for this particular argument which will be shown along with its notations to the user when
        help is triggered.

        - `parent`: If this argument is supposed to be registered under a specific group that is already registered using the `add_group`
        method, then specify the group name here.

        - `base`: This parameter is only valid when using `parent` parameter to specify a parent argument group. The `base` parameter
        specifies if the current argument is to be set as the group's base argument. If the base argument is present, only then all
        other arguments in that group is valid. Also, if any group's base argument is present but the group's parent (another group)'s
        base argument is not present, will raise an error. For more information regarding nested groups, refer documentation.

        - `default`: This is only valid when using capture mode as `'single'` or `'auto'` as for `'boolean'`, it will be automatically
        set to `False`. However, that `False` value can still be overridden if the default value is specified. This paramater holds no
        value when the `required` parameter is set to True.

        - `required`: This parameter sets an internal mechanism which makes sure that the current argument is mandatorily present in the
        argument pool else generates an exception. This works differently for grouped arguments: if the group's base argument is marked
        as a required argument, it acts similar to a single argument being marked required, however, all other arguments if marked
        required, the requirement mechanism is only triggerd if the base argument is present in the argument pool, else doesn't matter.

        This is the sole method that has to be used to create arguments. If `object` method is used to access underlying `singlet` or
        `singlets` object to create arguments, it wont be parsed as intended.
        """


    def add_group(self, name: str, parent: str = ..., helptext: str = ...) -> Self:
        """Create an empty argument group with given parameters.
        
        Parameters:

        - `name`: name uniquely identifies the group and is the only value that can be used to
        access it.

        - `parent`: This parameter accepts another group's name under which this new group will be created.
        If left untouched, the group will be created independently at topmost level.

        - `helptext`: The helptext description to show users along with all arguments' helptexts under this group when
        help is triggered.

        This is the sole method that has to be used to create argument groups. If `object` method is used to access underlying
        `singlets` object to create another sub-group, it wont be parsed as intended.
        """

    @overload
    def parse(self) -> Dict[str, Any]:
        """Parse all arguments and argument groups and return their parsed value in the form of a dictionary where,
        the keys are names of objects and values are their parsed value from the argument pool. All groups will have
        another dictionary as their value, where the keys will be the names of arguments/groups under that group and
        so on."""


    @overload
    def parse(self, order: Union[str, Tuple[str, ...]], /) -> Any:
        """Parses a single argument or group. If the motive is to parse an argument no matter how deep into the object tree,
        just mention the name of the argument (str), if it is a group, use a tuple where the first element is the group name,
        followed by the order in which the contents are inteded to be parsed and returned. The order in which the group needs
        to be parsed must contain atleast one element.

        Example:

        ```python
        value = judge_instance.parse('argument1')
        value2 = judge_instance.parse(('arg_grp', 'arg_grp_arg1', 'arg_grp_arg2'))
        ```
        
        If parsing an argument, may return bool (if capture mode is boolean), str (if capture mode is single), tuple[str, ...]
        (if the capture mode is auto, however, if only one value is found, returns str) or the default value set for the argument
        if not present. If no default value is present, and the argument is not passed in the argument pool, returns None.

        If parsing a group, A tuple will be returned with contents in the order as specified, where each individual argument will
        follow the above mentioned return types and any sub-groups will just repeat this.
        """


    @overload
    def parse(self, *order: Union[str, Tuple[str, ...]]) -> Union[Tuple[Any, ...], Any]:
        """Parse any number of objects all at once in the given order.
        
        For arguments, just use their names to specify them, but for argument groups, use a tuple where the first element is the
        group name followed by the order of objects inside that group. If the group has subgroups, use the same tuple format for
        the same. The group construct requires atleast one entry after the group name.

        Example:
        
        ```python
        value = judge_instance.parse('arg1', 'arg2', ('grp1',)) # not valid, expects atleast one value for order after the group name.
        value = judge_instance.parse('arg1', ('grp1', 'grp1-arg1'), 'arg2') # valid
        value = judge_instance.parse('arg1', ('grp1', ('grp2', 'grp2-arg1'))) # valid
        ```
        """